import logging

import pandas as pd
from mltraq import Bunch

from cumulative.opts import options
from cumulative.plotting import Plot
from cumulative.transforms.frame.cluster import Cluster
from cumulative.transforms.frame.copy import Copy
from cumulative.transforms.frame.drop import Drop
from cumulative.transforms.frame.sample import Sample
from cumulative.transforms.frame.score import Score
from cumulative.transforms.frame.sort import Sort
from cumulative.transforms.row.apply import Apply
from cumulative.transforms.row.bin import Bin
from cumulative.transforms.row.cumsum import CumSum
from cumulative.transforms.row.diff import Diff
from cumulative.transforms.row.interpolate import Interpolate
from cumulative.transforms.row.scale import Scale
from cumulative.transforms.transform import Transform
from cumulative.utils.frames import columns_with_prefix
from cumulative.utils.lineage import Lineage
from cumulative.utils.validate import Validate

log = logging.getLogger(__name__)


class Cumulative:
    def __init__(self, df: pd.DataFrame):
        """
        Instantiate a new Cumulative interface to dataframe `df`.
        """

        self.df = df

        # Container for dataset metadata
        self.meta = Bunch()

        # Transforms
        self.scale = Scale(self)
        self.diff = Diff(self)
        self.cumsum = CumSum(self)
        self.interpolate = Interpolate(self)
        self.score = Score(self)
        self.copy = Copy(self)
        self.sort = Sort(self)
        self.cluster = Cluster(self)
        self.bin = Bin(self)
        self.drop = Drop(self)
        self.apply = Apply(self)
        self.sample = Sample(self)

        # Utilities
        self.plot = Plot(self)
        self.lineage = Lineage()
        self.explain = self.lineage.explain
        self.validate = Validate(self)
        self.check = self.validate.check

    def register_transform(self, name: str, cls: Transform):
        """
        Register a new transform named `name`, implemented by class `cls`.
        """
        self.__dict__[name] = cls(self)

    def frame(self, src: str | None = None) -> pd.DataFrame:
        """
        Return dataframe with `src` as column prefix.
        """

        src = options().get("transforms.src", prefer=src)
        return self.df[columns_with_prefix(self.df, src)]

    def describe(self, src: str | None = None):
        """
        Print basic statistics about the collection in the `src` dimension.
        """

        src = options().get("transforms.src", prefer=src)

        def min_max_diff(values):
            return f"min={values.min()} max={values.max()} diff={values.max() - values.min()}"

        print(f"Count...: {len(self.df)}")
        print(f"Length..: {min_max_diff(self.df[f'{src}.x'].apply(len))}")
        print(f"X min...: {min_max_diff(self.df[f'{src}.x'].apply(min))}")
        print(f"X max...: {min_max_diff(self.df[f'{src}.x'].apply(max))}")
        print(f"Y min...: {min_max_diff(self.df[f'{src}.y'].apply(min))}")
        print(f"Y max...: {min_max_diff(self.df[f'{src}.y'].apply(max))}")



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


def cumsum(values: np.ndarray, prefix: str) -> dict:
    """
    Return dictionary with cumulative sum of `values`, nested in key `prefix`.
    """
    return {f"{prefix}": np.cumsum(values)}


class CumSum(Transform):
    def transform_row(self, row: pd.Series, src: str, kind: str = "y") -> pd.Series:
        """
        Apply cumulative sum to `kind` dimension(s) ("x", "y", or "xy") from
        prefix `src`.
        """

        if kind == "xy":
            x = cumsum(row[f"{src}.x"], "x")
            y = cumsum(row[f"{src}.y"], "y")
        elif kind == "x":
            x = cumsum(row[f"{src}.x"], "x")
            y = {"y": row[f"{src}.y"]}
        elif kind == "y":
            x = {"x": row[f"{src}.x"]}
            y = cumsum(row[f"{src}.y"], "y")

        return pd.Series({**x, **y})



import pandas as pd

from cumulative.transforms.transform import Transform


class Sample(Transform):
    def apply(self, src: str, n: int = None, m: int = None, frac: float = None) -> pd.DataFrame:
        """
        Sample dataset, either with a fixed number `n` of rows, or a
        percentage `frac`. If `m` is provided, up to `m` rows are considered.
        """

        if m:
            n = min(m, len(self.c.df))

        self.c.df = self.c.df.sample(frac=frac, n=n, random_state=123)



import numpy as np

from cumulative.utils.warnings import warn


class ValidationWarning(UserWarning):
    pass


class Validate:
    def __init__(self, c):
        self.c = c

    def check(self, prefix: str = None, deep=True):  # noqa: C901
        """
        Verify that linked cumulative data frame doesn't countain nans or infs,
        that there are both rows and columns.
        If `deep` is True, also arrays within cells are checked.
        If issues are detected, a warning is generated, with `prefix` string.
        """

        df = self.c.df

        if prefix:
            prefix = "[" + prefix + "] "
        else:
            prefix = ""

        # Check for invalid values (nans, infs)
        count_rows = len(df)
        count_invalid_rows = df.isin([np.inf, -np.inf, np.nan]).any(axis=1).sum()
        if count_invalid_rows > 0:
            warn(
                f"{prefix}Invalid rows (nans or infs): "
                f"{count_invalid_rows} ({count_invalid_rows / count_rows * 100:.0f}%)",
                category=ValidationWarning,
                stacklevel=1,
            )

        if deep and len(df) > 0:
            # Check arrays inside cells
            count_cols = 0
            count_cells = 0
            count_invalid_cols = 0
            invalid_cols = []
            count_invalid_cells = 0
            example_idx = None
            for col in df.columns:
                if isinstance(df.iloc[0][col], np.ndarray):
                    count_cols += 1
                    invalid = 0
                    for idx, row in df.iterrows():
                        count_cells += 1
                        if not np.all(np.isfinite(row[col])):
                            invalid += 1
                            if not example_idx:
                                example_idx = idx
                    if invalid > 0:
                        count_invalid_cells += invalid
                        count_invalid_cols += 1
                        invalid_cols.append(col)

            if count_invalid_cols > 0:
                warn(
                    f"{prefix}Invalid cells (nans or infs) in {count_invalid_cols} columns {invalid_cols}: "
                    f"{count_invalid_cells} ({count_invalid_cells / count_cells * 100:.0f}%), "
                    "sample idx: {example_idx}",
                    category=ValidationWarning,
                    stacklevel=1,
                )

        # Check for duplicate column names
        if len(set(df.columns)) != len(df.columns):
            warn(f"{prefix}Duplicate column names", category=ValidationWarning, stacklevel=1)

        if len(df.columns) == 0:
            warn(f"{prefix}No columns", category=ValidationWarning, stacklevel=1)

        if len(df) == 0:
            warn(f"{prefix}No rows", category=ValidationWarning, stacklevel=1)



import logging
from contextlib import _GeneratorContextManager, contextmanager
from typing import Any, Dict, Optional

import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from cumulative.opts import options

log = logging.getLogger(__name__)


def default_subplots() -> tuple[Figure, Axes]:
    """
    Create subplots with a given figsize config via options.
    """
    fig, ax = plt.subplots(figsize=options().get("plot.ctx.default.figsize"))
    return fig, ax


@contextmanager
def plot_notemplate_ctx(  # noqa: C901
    x_label: Optional[str] = None,
    y_label: Optional[str] = None,
    legend: Optional[Any] = None,
    interactive_mode: Optional[bool] = None,
    backend: Optional[str] = None,
    title: Optional[str] = None,
    facecolor: Optional[str] = None,
    x_lim: Optional[Dict] = None,
    y_lim: Optional[Dict] = None,
    spines_bottomleft: bool = False,
    x_minor_locator: Optional[float] = None,
    x_major_locator: Optional[float] = None,
    y_minor_locator: Optional[float] = None,
    y_major_locator: Optional[float] = None,
    y_grid: bool = False,
    hatches: bool = False,
    ax: Optional[Any] = None,
    figsize: Optional[tuple[int, int]] = None,
    show_axes: bool = True,
    show: bool = True,
    savefig_svg: Optional[str] = None,
    close: bool = True,
    mpl_rc: Optional[str] = None,
    mpl_style: str = "default",
):
    """
    Prepare a matplotlib plot (single sub-plot):

    `x_label`: X label
    `y_label`: Y label
    `legend`: Dictionary passed to ax.legend(...)
    `title`: Title of the plot
    `interactive_mode`: set plt.ioff() or plt.ion()
    `backend`: Backend to use
    `facecolor`: Color of drawing area
    `yerr`: If true, report Y error bars
    `x_lim`: X limits, passed to ax.set_xlim(...)
    `y_lim`: Y limits, passed to ax.set_ylim(...)
    `spines_bottomleft`: Show only left,bottom spines
    `x_minor_locator`: X Minor locator
    `x_major_locator`: X Major locator
    `y_minor_locator`: Y Minor locator
    `y_major_locator`: Y Major locator
    `y_logscale`: If true, set logscale for Y
    `y_grid`: If true, show grid on Y
    `hatches`: Show hatches on bars
    `ax`: If not None, use it as axis object to draw on
    `fixsize`: size (x,y) of the subplot
    `show`: If true, call plt.show()
    `show_axes`: If false, empty white canvas without axes
    `savefig_svg`: Save figure to pathname in SVG format
    `close`: If true, close figure as last step
    `mpl_rc`': matplotlib rc options
    `mpl_style`: matplotlib style options
    """

    rc = options().get("matplotlib.rc", prefer=mpl_rc, otherwise={})
    style = options().get("matplotlib.style", prefer=mpl_style, otherwise={})

    current_backend = mpl.get_backend()
    current_interactive_mode = plt.isinteractive()

    if savefig_svg:
        # If saving figure to file, don't display it (we cannot render to both.)
        show = False

    if backend:
        plt.switch_backend(backend)

    if interactive_mode is not None:
        if interactive_mode is True:
            plt.ion()
        else:
            plt.ioff()

    with plt.rc_context(rc), plt.style.context(style):
        if ax is None:
            # Note: You can retrieve the figure with ax.get_figure().
            fig, ax = plt.subplots(figsize=figsize if figsize else (3, 3))

        try:
            yield ax
        finally:
            if facecolor:
                ax.set_facecolor(facecolor)

            if spines_bottomleft:
                ax.spines["top"].set_visible(False)
                ax.spines["right"].set_visible(False)

            if not show_axes:
                ax.set_axis_off()

            if x_minor_locator:
                ax.yaxis.set_minor_locator(mtick.MultipleLocator(x_minor_locator))
            if x_major_locator:
                ax.yaxis.set_major_locator(mtick.MultipleLocator(x_major_locator))
            if y_minor_locator:
                ax.yaxis.set_minor_locator(mtick.MultipleLocator(y_minor_locator))
            if y_major_locator:
                ax.yaxis.set_major_locator(mtick.MultipleLocator(y_major_locator))

            ax.set_xlabel(x_label)
            ax.set_ylabel(y_label)
            ax.tick_params(axis="x", labelrotation=0)

            if hatches:
                hatches = ("./O\o*" * len(ax.patches))[: len(ax.patches)]
                for idx, bar in enumerate(ax.patches):
                    bar.set_hatch(hatches[idx])

            if x_lim:
                ax.set_xlim(**x_lim)
            if y_lim:
                ax.set_ylim(**y_lim)

            if y_grid:
                ax.grid(axis="y", which="major")

            if title is not None:
                ax.set_title(title, fontsize=plt.rcParams["font.size"] * 1)

            if legend:
                if isinstance(legend, dict):
                    ax.legend(**legend)
                else:
                    ax.legend()

            if show:
                plt.show()

            if backend:
                plt.switch_backend(current_backend)

            if interactive_mode is not None:
                if current_interactive_mode is True:
                    plt.ion()
                else:
                    plt.ioff()

            if savefig_svg is not None:
                plt.savefig(savefig_svg, format="svg", bbox_inches="tight")

            if close:
                plt.close(fig)


def plot_ctx(template_name: Optional[str] = None, **params) -> _GeneratorContextManager:
    """
    Return a context manager generator with options pulled
    from a template name `template_name`, and custom ones via `params`.
    """
    if not template_name:
        template_name = "plot.ctx.default"
    params = options().get(template_name) | params
    return plot_notemplate_ctx(**params)


class Plot:
    """
    This class provides an interface to a curated set of visualization routines.
    """

    def __init__(self, c):
        """
        Initializes the plotting interface for the `c` Cumulative instance.
        """
        self.c = c

    def xrays(
        self,
        src: str | None = None,
        ax: mpl.axes.Axes | None = None,
        alpha: float = 1,
        ms: float = 1,
        lw: float = 1,
        k: int = 20,
        style: str = "-",
        color=None,
    ):
        """
        Interpolate series on `k` points and render them as monochrome points/curves.
        """

        src = options().get("transforms.src", prefer=src)
        tmp = options().get("transforms.tmp") + ".plot"

        with options().ctx({"transforms.src": tmp, "transforms.dst": tmp}):
            self.c.copy(src=src)
            self.c.interpolate(method="pchipp", k=k, num=k)
            self.c.plot.draw(ax=ax, alpha=alpha, ms=ms, lw=lw, style=style, color=color)
            self.c.drop()
            return self

    def draw(
        self,
        src: str | None = None,
        ax: mpl.axes.Axes | None = None,
        style: str = ".",
        ms: float = 2,
        lw: float = 1,
        score: str | None = None,
        alpha: str | None = 0.5,
        only_changes: bool = False,
        color=None,
        colormap=None,
    ):
        """
        Basic visualization of a collection of series.
        """

        if not ax:
            _, ax = default_subplots()
            force_show = True
        else:
            force_show = False

        src = options().get("transforms.src", prefer=src)

        color = options().get("plot.color", prefer=color)
        cmap = mpl.colormaps[options().get("plot.colormap", prefer=colormap)]

        for _, row in self.c.df.iterrows():
            row_color = cmap(row[score]) if isinstance(score, str) else color
            row_alpha = row[alpha] if isinstance(alpha, str) else alpha

            if only_changes:
                # If activated, only the first occurrence of a repeated value is retained
                # and the subsequent ones are set to nan (lossless compression, provided
                # there are no missing values in the original series.)
                a = row[f"{src}.y"].copy()
                a = np.where(np.insert(np.diff(a), 0, 1) != 0, a, np.nan)
            else:
                a = row[f"{src}.y"]

            pd.Series(a, index=row[f"{src}.x"]).plot(
                style=style,
                lw=lw,
                ms=ms,
                color=row_color,
                alpha=row_alpha,
                ax=ax,
            )

        if force_show:
            # If the axis is the default one, we force the rendering of the plot
            # before returning (otherwise, it remains in the queue of plots to display.)
            plt.show()

        return self

    def fingerprint(
        self, src: str | None = None, ax: mpl.axes.Axes | None = None, score: str = "base.z", style="-", alpha=0.5
    ):
        """
        Sample and interpolate series on max 100 series, 100 points each, and use the ".z" suffix to select the
        color from the default colormap. This visualization is robust to large datasets, as it samples and
        simplifies the curves.
        """

        src = options().get("transforms.src", prefer=src)
        tmp = options().get("transforms.tmp") + ".plot"

        with options().ctx({"transforms.src": tmp, "transforms.dst": tmp}):
            # Copy from src to tmp, and then work only on tmp (the default src/dst)
            self.c.copy(src=src)
            self.c.sample(m=100)
            self.c.interpolate(method="pchipp", k=100, num=100)
            # Transform .interpolate cleans the tmp prefix, this is why we must create the score column afterwards.
            self.c.score(src=score, dst=f"{tmp}.score", method="value")
            # The highest value ir rendered last, increasing odds it is visible.
            self.c.sort(by=f"{tmp}.score.value")

            self.draw(src=src, ax=ax, style=style, ms=1, alpha=alpha, score=f"{tmp}.score.value", only_changes=True)
            # Transform .drop removes all columns with tmp prefix
            self.c.drop()
            return self



from cumulative.transforms.transform import Transform


class Sort(Transform):
    def apply(self, src, by: str, ascending: bool = True) -> None:
        """
        Sort rows in data frame based on column `by`, honoring
        `ascending` option. The order of the rows is affected.
        No new columns are created.
        """
        self.c.df = self.c.df.sort_values(by=by, ascending=ascending)



import inspect

from cumulative.opts import options


def get_kwargs() -> dict:
    """
    Return kwargs passed to funcion calling get_kwargs().
    """

    frame = inspect.currentframe().f_back
    keys, _, _, values = inspect.getargvalues(frame)
    kwargs = {}
    for key in keys:
        if key != "self":
            kwargs[key] = values[key]
    return kwargs


class Lineage:
    """
    Handling of lineage of operations.
    """

    def __init__(self):
        self.steps = []

    def track(self, name: str, dst: str | None = None, kwargs: dict | None = None):
        """
        Track transform call `name` to `dst` with `kwargs`.
        """

        if kwargs is None:
            kwargs = {}

        kwargs = kwargs.copy()
        src = options().get("transforms.src", prefer=kwargs.pop("src", None))
        dst = options().get("transforms.dst", prefer=kwargs.pop("dst", dst))
        drop = options().get("transforms.drop", prefer=kwargs.pop("drop", None))

        self.steps.append({"name": name, "src": src, "dst": dst, "drop": drop, "kwargs": kwargs})

    def clear(self):
        """
        Clear the currently tracked history.
        """

        self.steps = []

    def explain(self, max_len: int = 40, reverse: bool = False):
        """
        Print operations applied to data frame.
        """

        steps = list(enumerate(self.steps))

        if reverse:
            steps = steps[::-1]

        for idx, transform in steps:
            name = transform["name"]

            def stringify(v, string_delim: bool = False):
                if isinstance(v, str) and string_delim:
                    v = f'"{v}"'
                elif callable(v):
                    v = f"{v.__name__}"
                v = str(v)
                if len(v) > max_len:
                    v = v[:max_len] + ".."
                return v

            d = "" if transform["drop"] else "(no drop)"

            kwargs = ", ".join([f"{k}={stringify(v,string_delim=True)}" for k, v in transform["kwargs"].items()])
            if idx > 0:
                srcdst = f"{stringify(transform['src'])} -> {stringify(transform['dst'])} {d}"
            else:
                srcdst = f"-> {stringify(transform['dst'])} {d}"

            print(f"[{idx}] {name}({kwargs}): {srcdst}")



import logging

import numpy as np
import pandas as pd
from scipy.interpolate import PchipInterpolator

from cumulative.opts import options
from cumulative.transforms.transform import Transform
from cumulative.utils.warnings import warn

log = logging.getLogger(__name__)


class InterpolationWarning(UserWarning):
    pass


class ExceptionInvalidMethod(Exception):
    pass


def interpolate_interp(x: np.ndarray, y: np.ndarray, k: int = 2) -> tuple[callable, dict]:
    """
    Interpolate x,y curve with `k` equidistant points using one-dimensional linear
    interpolation for monotonically increasing sample points.
    """

    # Percentile-based Linear Interpolation (PLI)
    xp = np.linspace(x[0], x[-1], num=k)
    yp = np.interp(xp, x, y)

    def m(x):
        return np.interp(x, xp, yp)

    attrs = {"model.name": "interp", "model.params": (xp, yp), "model.size": k * 2}

    return m, attrs


def interpolate_pchip(x: np.ndarray, y: np.ndarray) -> tuple[callable, dict]:
    """
    PCHIP 1-D monotonic cubic interpolation on x,y curve.
    The resulting interpolation passed thru all x,y points.
    """

    # possible source of warnings for PCHIP:
    # https://stackoverflow.com/questions/14461346/python-pchip-warnings

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    m = PchipInterpolator(x, y)
    attrs = {"model.name": "pchip", "model.size": len(x) * 2}
    return m, attrs


def interpolate_pchipp(x: np.ndarray, y: np.ndarray, k: int = 2, p: np.ndarray | None = None) -> tuple[callable, dict]:
    """
    PCHIP 1-D monotonic cubic interpolation on `k` equi-distant (or `p`) percentiles.
    1. First, we fit a PCHIP model with all data points (x,y).
    2. We then extract the (x_hat,y_hat) pairs at percentiles x_hat (with k: equi-distant).
    3. We fit a new PCHIP model on (x_hat,y_hat), and return it.
    """

    # Seed the random number generator for reproducibility
    np.random.seed(options().get("reproducibility.random_seed"))

    m = PchipInterpolator(x, y)
    if p is None:
        p = np.linspace(0, 1, k)
    x = x[0] + (x[-1] - x[0]) * p
    xp = x[0] + (x[-1] - x[0]) * p
    yp = m(xp)

    m = PchipInterpolator(xp, yp)
    attrs = {"model.name": "pchipp", "model.params.x": xp, "model.params.y": yp, "model.size": len(xp) * 2}

    return m, attrs


def interpolate(x: np.ndarray, y: np.ndarray, method: str = "interp", **method_kwargs) -> tuple[callable, dict]:
    """
    Interpolate x,y curve using `method` and its key arguments, returning a callable
    to generate Y values from an X array, and interpolation properties. In case of errors,
    an "error" key is also added, with an error message.
    """

    try:
        if method == "interp":
            m, attrs = interpolate_interp(x, y, **method_kwargs)
        elif method == "pchip":
            m, attrs = interpolate_pchip(x, y, **method_kwargs)
        elif method == "pchipp":
            m, attrs = interpolate_pchipp(x, y, **method_kwargs)
        else:
            raise ExceptionInvalidMethod(f"No valid method: '{method}'")
    except (RuntimeError, TypeError, ValueError) as e:
        error = f"{__name__}: {e.__class__.__name__}: {e}"
        warn(error, category=InterpolationWarning, stacklevel=1)
        attrs = {"model.name": method, "error": error}

        def m(x):
            return np.full(x.shape, np.nan)

    return m, attrs


class Interpolate(Transform):
    def transform_row(
        self, row: pd.Series, src: str, method: str = "interp", num: int = 1000, **method_kwargs
    ) -> pd.Series:
        """
        Interpolate  x,y arrays with prefix `src` using `method` and `num` data points.
        """
        func, attrs = interpolate(row[f"{src}.x"], row[f"{src}.y"], method, **method_kwargs)
        x = np.linspace(row[f"{src}.x"][0], row[f"{src}.x"][-1], num=num)
        y = func(x)
        z = func(row[f"{src}.x"])
        return pd.Series({"x": x, "y": y, "z": z} | attrs)



import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform


def bin_values(values: np.ndarray, num: int, prefix: str) -> dict:
    """
    Given `values`, bin them in `num` equidistant bins`.
    """
    bins = np.linspace(values[0], values[-1], num=num)
    values_hat = bins[np.searchsorted(bins, values)]
    return {f"{prefix}": values_hat}


class Bin(Transform):
    def transform_row(self, row: pd.Series, src: str, kind: str = "y", num: int = 10):
        """
        Simplify curve binning Y values on `num` bins. `kind` control the source
        of the values: "x", "y", or "xy" (both "x" and "y", independently.)
        """

        if kind == "xy":
            x = bin_values(row[f"{src}.x"], num, "x")
            y = bin_values(row[f"{src}.y"], num, "y")
        elif kind == "x":
            x = bin_values(row[f"{src}.x"], num, "x")
            y = {"y": row[f"{src}.y"]}
        elif kind == "y":
            x = {"x": row[f"{src}.x"]}
            y = bin_values(row[f"{src}.y"], num, "y")

        return pd.Series({**x, **y})



import pandas as pd

from cumulative import Cumulative


def load_twoseries() -> Cumulative:
    """
    Load a dataset of two time series.
    """

    df = pd.DataFrame(
        {
            "base.x": [[0, 1, 2, 3, 4, 5], [0, 1, 2, 3]],
            "base.y": [[10, 20, 30, 40, 50, 60], [5, 25, 15, 5]],
        }
    )

    c = Cumulative(df)

    c.lineage.track("load_twoseries")

    return c



import logging

import numpy as np
import pandas as pd

from cumulative.transforms.transform import Transform

log = logging.getLogger(__name__)


def scale(values: np.ndarray, prefix: str) -> dict:
    """
    Scale `values` to [0,1] range with `prefix` keys.
    """

    v_min = np.min(values)
    v_max = np.max(values)

    if v_min == v_max:  # minmax scaling not defined, defaulting to 1
        V = np.full(values.shape[0], 1.0)
    else:
        V = (values - v_min) / (v_max - v_min)
    return {f"{prefix}": V, f"{prefix}.min": v_min, f"{prefix}.max": v_max}


class Scale(Transform):
    def transform_row(self, row: pd.Series, src: str, kind: str = "y") -> pd.Series:
        """
        Apply minmax normalization to range [0,1] for `kind` dimensions ()"x", "y" or "xy").
        """

        if kind == "xy":
            x = scale(row[f"{src}.x"], "x")
            y = scale(row[f"{src}.y"], "y")
        elif kind == "x":
            x = scale(row[f"{src}.x"], "x")
            y = {"y": row[f"{src}.y"]}
        elif kind == "y":
            x = {"x": row[f"{src}.x"]}
            y = scale(row[f"{src}.y"], "y")

        return pd.Series({**x, **y})



import numpy as np
import pandas as pd

from cumulative.cumulative import Cumulative

names = [
    "ACSF1",
    "Adiac",
    "AllGestureWiimoteX",
    "AllGestureWiimoteY",
    "AllGestureWiimoteZ",
    "ArrowHead",
    "BME",
    "Beef",
    "BeetleFly",
    "BirdChicken",
    "CBF",
    "Car",
    "Chinatown",
    "ChlorineConcentration",
    "CinCECGTorso",
    "Coffee",
    "Computers",
    "CricketX",
    "CricketY",
    "CricketZ",
    "Crop",
    "DiatomSizeReduction",
    "DistalPhalanxOutlineAgeGroup",
    "DistalPhalanxOutlineCorrect",
    "DistalPhalanxTW",
    "DodgerLoopDay",
    "DodgerLoopGame",
    "DodgerLoopWeekend",
    "ECG200",
    "ECG5000",
    "ECGFiveDays",
    "EOGHorizontalSignal",
    "EOGVerticalSignal",
    "Earthquakes",
    "ElectricDevices",
    "EthanolLevel",
    "FaceAll",
    "FaceFour",
    "FacesUCR",
    "FiftyWords",
    "Fish",
    "FordA",
    "FordB",
    "FreezerRegularTrain",
    "FreezerSmallTrain",
    "Fungi",
    "GestureMidAirD1",
    "GestureMidAirD2",
    "GestureMidAirD3",
    "GesturePebbleZ1",
    "GesturePebbleZ2",
    "GunPoint",
    "GunPointAgeSpan",
    "GunPointMaleVersusFemale",
    "GunPointOldVersusYoung",
    "Ham",
    "HandOutlines",
    "Haptics",
    "Herring",
    "HouseTwenty",
    "InlineSkate",
    "InsectEPGRegularTrain",
    "InsectEPGSmallTrain",
    "InsectWingbeatSound",
    "ItalyPowerDemand",
    "LargeKitchenAppliances",
    "Lightning2",
    "Lightning7",
    "Mallat",
    "Meat",
    "MedicalImages",
    "MelbournePedestrian",
    "MiddlePhalanxOutlineAgeGroup",
    "MiddlePhalanxOutlineCorrect",
    "MiddlePhalanxTW",
    "MixedShapesRegularTrain",
    "MixedShapesSmallTrain",
    "MoteStrain",
    "NonInvasiveFetalECGThorax1",
    "NonInvasiveFetalECGThorax2",
    "OSULeaf",
    "OliveOil",
    "PLAID",
    "PhalangesOutlinesCorrect",
    "Phoneme",
    "PickupGestureWiimoteZ",
    "PigAirwayPressure",
    "PigArtPressure",
    "PigCVP",
    "Plane",
    "PowerCons",
    "ProximalPhalanxOutlineAgeGroup",
    "ProximalPhalanxOutlineCorrect",
    "ProximalPhalanxTW",
    "RefrigerationDevices",
    "Rock",
    "ScreenType",
    "SemgHandGenderCh2",
    "SemgHandMovementCh2",
    "SemgHandSubjectCh2",
    "ShakeGestureWiimoteZ",
    "ShapeletSim",
    "ShapesAll",
    "SmallKitchenAppliances",
    "SmoothSubspace",
    "SonyAIBORobotSurface1",
    "SonyAIBORobotSurface2",
    "StarLightCurves",
    "Strawberry",
    "SwedishLeaf",
    "Symbols",
    "SyntheticControl",
    "ToeSegmentation1",
    "ToeSegmentation2",
    "Trace",
    "TwoLeadECG",
    "TwoPatterns",
    "UMD",
    "UWaveGestureLibraryAll",
    "UWaveGestureLibraryX",
    "UWaveGestureLibraryY",
    "UWaveGestureLibraryZ",
    "Wafer",
    "Wine",
    "WordSynonyms",
    "Worms",
    "WormsTwoClass",
    "Yoga",
]


def load_ucr2018(pathdir: str = "./UCRArchive_2018", name: str = "Beef") -> Cumulative:
    """
    Load dataset from UCR Archive (2018) by passing `pathdir` as root directory
    of the archive and the dataset name `name`.
    """

    df1 = pd.read_csv(f"{pathdir}/{name}/{name}_TRAIN.tsv", sep="\t", header=None)
    df2 = pd.read_csv(f"{pathdir}/{name}/{name}_TEST.tsv", sep="\t", header=None)
    df = pd.concat([df1, df2], ignore_index=True)

    def f(row):
        d = {}
        d["name"] = row.name
        d["base.x"] = np.linspace(1, len(row) - 1, num=len(row) - 1, dtype=int)
        d["base.y"] = pd.Series([row[pos] for pos in d["base.x"]]).ffill().bfill().values
        d["base.z"] = row[0]
        return pd.Series(d)

    df = df.apply(f, axis=1)

    c = Cumulative(df)

    return c



import itertools

import numpy as np
import pandas as pd
from scipy.special import betainc

from cumulative import Cumulative
from cumulative.opts import options
from cumulative.utils.lineage import get_kwargs


def betainc_record(a: float, b: float, num: int, dst: str) -> dict:
    """
    Generate a new record representing the betainc series with parameters `a` and `b`
    using `num` equidistant points between 0 and 1, with `dst` key prefix.
    """

    x = np.linspace(0, 1, num)
    y = betainc(a, b, x)
    return {f"{dst}.a.a": a, f"{dst}.a.b": b, f"{dst}.x": x, f"{dst}.y": y}


def load_betainc(a_min=0.1, a_max=10, b_min=0.1, b_max=10, size=10, num=100, dst=None) -> Cumulative:
    """
    Load a synthetic dataset of `size` betainc curves with a and b parameters ranging within
    intervals [`a_min`, `a_max`] and [`b_min`, `b_max`], represented by `num` points each.
    """

    a = np.linspace(a_min, a_max, size)
    b = np.linspace(b_min, b_max, size)

    dst = options().get("transforms.dst", prefer=dst)

    records = []
    for ab in itertools.product(a, b):
        record = betainc_record(ab[0], ab[1], num, dst)
        records.append(record)

    df = pd.DataFrame(records)
    c = Cumulative(df)

    c.lineage.track("load_betainc", dst, get_kwargs())

    return c



import pandas as pd
from mltraq.utils.exceptions import ExceptionWithMessage


class NoMatchingColumns(ExceptionWithMessage):
    pass


def columns_with_prefix(df: pd.DataFrame, prefix: str, errors: str = "ignore") -> list[str]:
    """
    Return list of columns in `df` dataframe with `prefix`. If no matches and errors == "raise",
    raise NoMatchingColumns exception.
    """

    cols = [col for col in df.columns if col.startswith(f"{prefix}.") or col == prefix]
    if errors == "raise" and len(cols) == 0:
        raise NoMatchingColumns(f"No columns with prefix {prefix}")
    return cols


def drop_prefix(src, cols: list[str]) -> list[str]:
    return [col.removeprefix(f"{src}.") for col in cols]


def rows_with_prefix(s: pd.Series, prefix: str, errors="ignore") -> list[str]:
    """
    Return list of index values in `s` series with `prefix`. If no matches and errors == "raise",
    raise NoMatchingColumns exception.
    """

    cols = [col for col in s.index if col.startswith(f"{prefix}.") or col == prefix]
    if errors == "raise" and len(cols) == 0:
        raise NoMatchingColumns(f"No columns with prefix {prefix}")
    return cols


def drop_cols_with_prefix(df: pd.DataFrame, prefix: str, errors="ignore") -> pd.DataFrame:
    """
    Drop columns with `prefix` from `df` dataframe and return it.
    If no matches and errors == "raise", NoMatchingColumns is raised.
    """
    cols = columns_with_prefix(df, prefix, errors=errors)
    return df.drop(columns=cols)



import warnings

from cumulative.opts import options


def warn(*args, stacklevel=1, **kwargs):
    """
    Utility function to handle warnings.
    """
    if not options().get("warnings.disable"):
        warnings.warn(*args, stacklevel=stacklevel, **kwargs)



import warnings

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.exceptions import ConvergenceWarning

from cumulative.opts import options
from cumulative.transforms.transform import Transform


class Cluster(Transform):
    def apply(self, src: str, k: int = 2) -> pd.DataFrame:
        """
        At each `x` position, cluster `y` values with k-means.
        """

        # Seed the random number generator for reproducibility
        seed = options().get("reproducibility.random_seed")

        y = np.dstack(self.c.df[f"{src}.y"].values).squeeze()
        # `y`: first row contains y values for all series at idx=0

        y_hat = []

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ConvergenceWarning)

            for i in range(y.shape[0]):
                yi = y[i].reshape(-1, 1)
                kmeans = KMeans(n_clusters=k, random_state=seed, n_init="auto").fit(yi)
                yi_hat = kmeans.cluster_centers_[kmeans.predict(yi)].squeeze()
                y_hat.append(yi_hat)

        y_hat = np.array(y_hat)
        # `y`: first row contains clustered y values

        return pd.DataFrame(
            {"x": self.c.df[f"{src}.x"], "y": pd.Series(y_hat.T.tolist(), index=self.c.df.index)}, copy=True
        )
